from aiohttp import web
from typing import Dict
from ..parsing import RequestParser


def setCurrentUser(response: web.Response, user: Dict) -> web.Response:
    return response


# class CurrentUserField():


class AuthenticatedRequestParser(RequestParser):
    currentUser = None



from dataclasses import dataclass, field
from typing import Any, List, Dict

from ..errors import ParseError

import jwt
import bson
import traceback
import inspect
import string
import enum


class FieldTypes(str, enum.Enum):
    STRING = "string"
    BOOL = "boolean"
    OBJECT = "object"
    INTEGER = "int"
    FLOAT = "float"


@dataclass
class Field():
    optional: bool = False
    missingStatusCode: int = None
    default: Any = None
    exampleValues: List[Any] = field(default_factory=list)
    allowedTypes: List[str] = field(default_factory=list)
    description: str = ""
    parseFrom: str = None
    serializeTo: str = None

    def parser(self, value):
        return value

    def normalizer(self, value):
        return value

    def parse(self, value):
        value = self.parser(value)
        return self.normalize(value)

    def normalize(self, value):
        return self.normalizer(value)

    def serialize(self, value):
        return value

    def describe(self):
        return {
            "allowedTypes": self.allowedTypes,
            "optional": self.optional,
            "default": self.default,
            "description": self.description,
            "exampleValues": self.exampleValues,
        }

    def __repr__(self):
        return "FIELD"


@dataclass
class JwtField(Field):
    secretKey: str = ""

    def parser(self, value):
        value = super().parser(value)
        if not value:
            return None
        parsed = jwt.decode(
            value,
            self.secretKey,
            algorithms=['HS256'],
        )
        return parsed



def withSession(f):
    async def wrapper(*args, db=None, session=None, **kwargs):
        if not session:
            async with await db.client.start_session() as session:
                async with session.start_transaction():
                    return await f(*args, db=db, session=session, **kwargs)

        return await f(*args, db=db, session=session, **kwargs)

    return wrapper


from aiohttp import web
from typing import Dict, List

from .fields import Field, ListField, ErrorField
from .errors import ParseError
from ..errors import ClientError, ErrorList
import datetime
import inspect

from . import json


class DictParser():
    def __init__(self, settings=None, *args, **kwargs):
        self.settings = settings or {}
        return

    async def _parse_body(self, request):
        try:
            return json.loads(await request.text())
        except Exception as e:
            print(e)
            return {}

    def _getFields(self):
        return {
            f: theField
            for (f, theField) in filter(
                lambda k: isinstance(k[1], Field),
                inspect.getmembers(self),
            )
        }

    async def parse(self, source: Dict):
        parameters = {}
        errors = []
        for name, parameter in self._getFields().items():
            try:
                if name not in source:
                    if (not parameter.optional):
                        raise ParseError(
                            message="Missing required field: {}".format(name),
                            statusCode=parameter.missingStatusCode,
                        )
                    else:
                        parameters[name] = parameter.default
                else:
                    parameters[name] = parameter.parse(
                        source.get(name, parameter.default)
                    )
            except Exception as err:
                errors.append(err)
        if errors:
            raise ErrorList(errors)
        return parameters

    def describe(self):
        return {
            key: field.describe()
            for (key, field) in self._getFields().items()
        }


class RequestParser():
    BodyParser = DictParser
    QueryParser = DictParser
    CookiesParser = DictParser
    HeadersParser = DictParser
    UrlParser = DictParser

    def __init__(self, *args, **kwargs):
        self.settings = kwargs.get("settings", None)
        return

    async def _parse_body(self, request):
        try:
            if request.can_read_body:
                return await request.json()
            return None
        except Exception as e:
            print(e)
            raise ParseError(message="Unable to parse body")

    async def parse(self, request):
        parameters = {}
        errors = []

        if self.BodyParser:
            body = (await self._parse_body(request)) or {}
            try:
                parameters.update(await self.BodyParser().parse(body))
            except ErrorList as errs:
                errors += errs.errors

        if self.QueryParser:
            try:
                parameters.update(
                    await self.QueryParser().parse(request.query)
                )
            except ErrorList as errs:
                errors += errs.errors

        if self.HeadersParser:
            try:
                parameters.update(
                    await self.HeadersParser().parse(request.headers)
                )
            except ErrorList as errs:
                errors += errs.errors

        if self.CookiesParser:
            try:
                parameters.update(
                    await self.CookiesParser(settings=self.settings, ).parse(
                        request.cookies
                    )
                )
            except ErrorList as errs:
                errors += errs.errors

        try:
            parameters.update(
                await self.UrlParser(settings=self.settings, ).parse(
                    request.match_info
                )
            )

        except:
            raise ClientError(statusCode=404)

        if errors:
            raise ErrorList(errors)

        return parameters

    def describe(self):
        """
        Return a dictionary of possible parameters to be parsed
        """
        description = {
            "body": self.BodyParser().describe(),
            "url": self.UrlParser().describe(),
            "query": self.QueryParser().describe(),
        }

        for (k, v) in list(description.items()):
            if not v:
                del description[k]

        return description


class ResponseSerializer():
    def __init__(self, *args, **kwargs):
        return

    def _getFields(self):
        return {
            f: self.__getattribute__(f)
            for f in dir(self)
            if isinstance(self.__getattribute__(f), Field)
        }

    def makeResponse(self, body, statusCode, headers):
        """
        Convert a parsed body, statusCode, and headers into
        an aiohttp.web.Response
        """
        if body != None:
            response = web.json_response(
                body,
                status=statusCode,
                headers=headers or {},
            )
        else:
            response = web.Response(
                None,
                status=statusCode,
                headers=headers or {},
            )

        return response

    def parseResult(self, result):
        """
        Walk the list of fields and convert them into a(stinrg)
        dictionary ready to be passed to self.makeResponse
        """
        body = {}
        fields = self._getFields()
        for name, resultField in fields.items():

            if hasattr(resultField, "default"):
                value = result.get(name, resultField.default)
            else:
                value = result[name]

            try:
                value = resultField.serialize(value)
            except Exception as e:
                print(e)
                pass

            serializeTo = getattr(resultField, "serializeTo", name) or name
            body[serializeTo] = value
        return body

    def writeCookies(
        self,
        response,
        cookies=None,
        deletedCookies=None,
    ):
        """
        Add/delete cookies to/from a response
        """
        for c in (cookies or {}).values():
            try:
                value = c["value"].decode("utf")
            except:
                value = c["value"]

            kwargs = {
                "name": c["name"],
                "value": value,
                "secure": c["secure"],
                "httponly": c["httpOnly"],
                "expires": c.get(
                    "expires",
                    datetime.datetime.utcnow() + datetime.timedelta(days=1)
                )
            }

            if c.get("domain", None):
                kwargs["domain"] = c["domain"]

            response.set_cookie(**kwargs)

        for c in (deletedCookies or {}).values():
            response.del_cookie(c["name"], domain=c.get("domain", None))

        return response

    async def serialize(
        self,
        result,
        statusCode=200,
        headers=None,
        cookies=None,
        deletedCookies=None,
        dropBody=False,
    ):
        body = self.parseResult(result)
        headers = headers or {}
        headers = {k: "; ".join(vs) for k, vs in headers.items()}

        response = self.makeResponse(
            body=(body if not dropBody else None),
            statusCode=statusCode,
            headers=headers,
        )

        response = self.writeCookies(
            response=response,
            cookies=cookies,
            deletedCookies=deletedCookies,
        )

        return response

    def describe(self):
        return {
            "body": {
                key: field.describe()
                for (key, field) in self._getFields().items()
            }
        }


class ResponseSerializerWithErrors(ResponseSerializer):
    errors = ListField(ErrorField)


class HtmlResponseSerializer(ResponseSerializer):
    def __init__(self, render=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.render = render

    def parseResult(self, result):
        if self.render:
            return self.render(result)
        return result

    def makeResponse(self, body, statusCode, headers):
        return web.Response(
            body,
            status=statusCode,
            headers=headers,
        )


def handleAcceptHeader(
    header: str,  # the header value
    serverTypes: List[str],
) -> str:
    """
    Given the value of an HTTP request Accept header and a list
    of possible types the server supports, and any requested parameters
    """
    clientTypes = [h.strip() for h in header.split(",")]
    clientTypesWithParameters = []
    for ii, t in enumerate(clientTypes[::-1]):
        contentType, *parameters = t.split(";")
        contentType = contentType.strip().lower()

        parameters = [p.strip() for p in parameters]
        parsedParameters = {}
        quality = 1.0
        for p in parameters:
            key, *value = p.split("=")
            value = "=".join(value).strip()
            if key == "q":
                try:
                    quality = float(value)
                except:
                    raise ParseError(
                        message="Invalid Accept Parameter, q must be a number",
                    )
            parsedParameters[key] = value

        clientTypesWithParameters.append({
            "contentType": contentType,
            "quality": (quality, ii),
            "parameters": parsedParameters,
        })

    clientTypesWithParameters = sorted(
        clientTypesWithParameters,
        key=lambda k: k["quality"],
        reverse=True,
    )

    for t in clientTypesWithParameters:
        contentType = t["contentType"]
        parameters = t["parameters"]
        if contentType == "*/*":
            return serverTypes[0], parameters
        elif contentType.endswith("/*"):
            for serverType in serverTypes:
                if serverType.startswith(contentType[:-1]):
                    return serverType, parameters

        elif contentType in serverTypes:
            return contentType, parameters

    return None, None


import motor
import asyncio
import bson
import datetime

from pymongo.errors import DuplicateKeyError
from ..errors import ClientError
from ..errors.mongo import parseDuplicateKeyError

from ..db.transactions import withSession
from ..parsing.fields import (
    ObjectIdField,
    StringField,
    DictField,
    DatetimeField,
)


async def createActivityCollection(db):
    try:
        await db.create_collection("activity")
    except:
        pass
    await db.command(
        "collMod",
        "activity",
        validator={
            "$jsonSchema": {
                "bsonType": "object",
                "required": [
                    "action", "actorID", "targetIDs", "targetType", "payload"
                ],
                "properties": {
                    "action": {
                        "bsonType": "string",
                    },
                    "actorID": {
                        "bsonType": "objectId",
                    },
                    "targetIDs": {
                        "bsonType": "array",
                        "items": {
                            "bsonType": "objectId",
                        },
                    },
                    "targetType": {
                        "bsonType": "string",
                    },
                    "payload": {
                        "bsonType": "object",
                    },
                }
            }
        }
    )

    await asyncio.wait([
        db.activity.create_index([("actorID", 1)], ),
        db.activity.create_index([("targetIDs", 1)], ),
    ])


@withSession
async def createActivity(
    db,
    action: str,
    actorID: bson.ObjectId,
    targetIDs,
    targetType: str,
    payload: dict,
    session=None,
):
    activity = {
        "action": action,
        "actorID": actorID,
        "targetIDs": targetIDs,
        "targetType": targetType,
        "payload": payload,
    }
    try:
        result = await db.activity.insert_one(
            activity,
            session=session,
        )
        activity["_id"] = result.inserted_id

        return activity

    except Exception as err:
        raise err



import markdown

from aiohttp import web
from inspect import getdoc

from .method import Method, newMethod
from ..parsing import handleAcceptHeader, json
from ..documentation import requestDocumentationToApiBlueprint


class Endpoint():
    def __init__(
        self,
        name: str,
        settings=None,
        path: str = None,
    ):
        self.settings = settings or {}
        self.Options = self.generateOptionsClass(
            list(self.methods().keys()) + ["options"],
            settings=self.settings,
        )

        if hasattr(self, "Get"):
            # print(f"Should build HEAD route for {self}")
            self.Head = self.generateHeadClass(
                self.Get,
                settings=self.settings,
            )

        self.name = name
        self.path = path

    def methods(self):
        return {
            f.lower(): self.__getattribute__(f)
            for f in dir(self)
            if isinstance(self.__getattribute__(f), Method.__class__)
            and f[0] != "_"
        }

    async def handle(self, request: web.Request) -> web.Response:
        method = self.methods().get(request.method.lower(), None)
        if method:
            return await method(settings=self.settings).handle(request)
        else:
            return await self.methodNotAllowedHandler(method)

    async def methodNotAllowedHandler(self, method):
        return web.json_response(
            {"errors": ["Method not allowed"]},
            status=405,
        )

    async def handleDocumentation(
        self, request: web.Request, *args
    ) -> web.Response:
        docs = self.describe()
        acceptHeader = request.headers.get("Accept")

        if acceptHeader:
            contentType, _ = handleAcceptHeader(
                acceptHeader,
                [
                    "application/json",
                    "text/html",
                ],
            )
        else:
            contentType = "application/json"

        if contentType == "text/html":
            md = requestDocumentationToApiBlueprint(docs)
            htmlContent = markdown.markdown(md)
            html = f"<html><body>{htmlContent}</body></html>"

            return web.Response(
                text=html,
                headers={"Content-Type": "text/html"},
            )

        else:
            return web.Response(
                text=json.dumps(docs),
                headers={"Content-Type": "application/json"},
            )

    def describe(self):
        docs = {
            "description": getdoc(self),
        }

        if self.name:
            docs["name"] = self.name

        if self.path:
            docs["path"] = self.path

        docs["methods"] = {
            methodName: methodClass("").describe()
            for methodName, methodClass in self.methods().items()
        }
        return docs

    def getTests(self):
        tests = []
        for method in self.methods().values():
            tests += method(settings=self.settings).getTests()
        return tests

    def generateOptionsClass(self, methods, settings=None):
        methodString = ", ".join([m.upper() for m in methods])

        settings = settings or {}

        async def process(self, **kwargs):
            self.setHeader("Allow", methodString)

            # CORS
            self.setHeader("Access-Control-Request-Method", methodString)
            return

        name = getattr(self, 'name', 'Unknown Handler')

        return newMethod(
            name="Options",
            httpMethod="OPTIONS",
            description=
            f"Returns possible methods and content types for this endpoint.",
            process=process,
        )

    def generateHeadClass(self, getMethod, settings):
        class Head(Method):
            """
            HEAD Handler
            """
            name = "Head"
            description = "Returns the same response as a get, but with no body."
            Parser = getMethod.Parser
            process = getMethod.process

            async def serialize(self, result):
                return await self.Serializer().serialize(
                    result,
                    statusCode=self.statusCode,
                    headers=self.headers,
                    cookies=self.cookies,
                    deletedCookies=self.deletedCookies,
                    dropBody=True,
                )

        return Head



import faker
import sys
import unittest
import asyncio

sys.path.append(".")
fake = faker.Faker()

from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

from .. import Clericus
from ..config import defaultSettings, connectToDB
from ..schemas import createCollections
from ..routes import authentication as authenticationRoutes
from ..parsing import json


def async_test(f):
    def wrapper(self):
        return asyncio.run(f(self))

    return wrapper


class ClericusTestCase(AioHTTPTestCase):
    async def tearDownAsync(self):
        await self.db.client.drop_database(self.db.name)

    async def get_application(self) -> Clericus:
        settings = defaultSettings()
        settings["db"]["name"] = f"test{type(self).__name__}"
        settings = connectToDB(settings)
        await settings["db"].client.drop_database(settings["db"].name)
        self._settings = settings
        self.db = settings["db"]
        await createCollections(self.db)
        return Clericus(self._settings, logging=False)


class AuthenticatedClericusTestCase(AioHTTPTestCase):
    async def tearDownAsync(self):
        await self.db.client.drop_database(self.db.name)

    async def get_application(self) -> Clericus:
        settings = defaultSettings()
        settings["db"]["name"] = f"test{type(self).__name__}"
        settings = connectToDB(settings)
        await settings["db"].client.drop_database(settings["db"].name)
        self._settings = settings
        self.db = settings["db"]
        await createCollections(self.db)
        app = Clericus(self._settings, logging=False)

        app.addEndpoint(
            "/sign-up/",
            authenticationRoutes.SignUpEndpoint,
        )
        app.addEndpoint(
            "/log-in/",
            authenticationRoutes.LogInEndpoint,
        )
        app.addEndpoint(
            "/log-out/",
            authenticationRoutes.LogOutEndpoint,
        )
        app.addEndpoint(
            "/me/",
            authenticationRoutes.MeEndpoint,
        )

        return app

    async def login(self, user=None):
        if not user:
            user = {
                "username": fake.user_name(),
                "email": fake.email(),
                "password": fake.password(),
            }
            resp = await self.client.request(
                "POST",
                "/sign-up/",
                json=user,
            )
        else:
            resp = await self.client.request(
                "POST",
                "/log-in/",
                json=user,
            )
        return user

    async def logout(self):
        resp = await self.client.request(
            "GET",
            "/log-out/",
        )
        return None


import textwrap

from typing import List


def requestDocumentationToApiBlueprint(docs):
    queryParameters = []

    for method in docs.get("methods", {}).values():
        queryParameters += method.get(
            "requestParameters",
            {},
        ).get(
            "query",
            {},
        ).keys()

    queryParameterString = ",".join(sorted(set(queryParameters)))

    if queryParameters:
        queryParameterString = "{?" + queryParameterString + "}"

    s = f"""
# {docs.get("name", "Unnamed Route")} [{docs["path"]}{queryParameterString}]

{docs["description"]}
    """

    for method, data in docs["methods"].items():
        s += f"""\n## {data["name"]} [{method.upper()}]\n"""

        if data.get("description"):
            s += f"{data['description']}\n"

        parameters = data.get(
            "requestParameters",
            {},
        )

        displayableParameters = sorted(
            list(parameters.get(
                "url",
                {},
            ).items()) + list(parameters.get(
                "query",
                {},
            ).items())
        )

        bodyParameters = sorted(list(parameters.get(
            "body",
            {},
        ).items()))

        if displayableParameters:
            s += "\n+ Parameters\n"

            for name, parameter in displayableParameters:
                s += _parameterToMarkdown(
                    name=name,
                    parameter=parameter,
                )

        if bodyParameters:
            s += "\n Attributes\n"

            for name, parameter in bodyParameters:
                s += _parameterToMarkdown(
                    name=name,
                    parameter=parameter,
                )
        for example in data.get("examples", []):

            s += _testCaseToMarkdown(
                request=example["request"],
                response=example["response"],
                name="Unnamed",
            )

    return s


def _allowedTypesToString(allowedTypes: List) -> str:
    if len(allowedTypes) == 1:
        allowedTypesString = allowedTypes[0]
    elif allowedTypes:
        allowedTypesString = ",".join(allowedTypes)
    else:
        allowedTypesString = "any"
    return allowedTypesString


def _parameterToMarkdown(name: str, parameter: dict):
    s = ""
    allowedTypes = _allowedTypesToString(parameter.get("allowedTypes", []))

    optional = "optional" if parameter.get("optional") else "required"

    description = parameter.get("description") or ""

    default = parameter.get("default")

    s += f"\n\t+ {name} ({allowedTypes}, {optional}) - {description}\n"

    if default is not None:
        s += f"\t\t+ Default: ``{default}``\n"

    if parameter.get("allowedValues"):
        s += "\t\t+ Members\n"

        for value in parameter["allowedValues"]:
            s += f"\t\t\t+ `{value}`\n"

    return s


def _testCaseToMarkdown(request, response, name=""):
    contentType = request.headers.get("Content-Type", "application/json")

    s = f"+ Request {name} ({contentType})\n\n"
    if request.body:
        s += textwrap.indent(request.body, "\t") + "\n"

    contentType = request.headers.get("Content-Type", "application/json")
    s += f"+ Response {response.statusCode} ({contentType})\n\n"
    if response.body:
        s += textwrap.indent(response.body, "\t") + "\n"

    return s


from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict


@dataclass
class BoolField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.BOOL])

    def parser(self, value):
        try:
            return value == True
        except:
            raise ParseError(
                message="\"{}\" is not a valid boolean".format(value)
            )
        return value


import unittest
import asyncio
from aiohttp.test_utils import make_mocked_request

from ...handler import newMethod, Endpoint
from ...parsing import json
from ...parsing.fields import BoolField, StringField

from ..test_case import (
    ClericusTestCase,
    unittest_run_loop,
    async_test,
)


class TestMethods(unittest.TestCase):
    @async_test
    async def testMethodCreation(self):
        async def process(self):
            return {"ok": True, "text": "stuff"}

        method = newMethod(
            name="Test Method",
            httpMethod="Get",
            description="This is a test handler",
            process=process,
            responseFields={
                "ok": BoolField(),
                "text": StringField(),
            },
        )

        req = make_mocked_request('GET', '/')

        resp = await method().handle(req)
        respBody = json.loads(resp.body)

        self.assertEqual(respBody["ok"], True)
        self.assertEqual(respBody["text"], "stuff")


class TestUrlParameters(ClericusTestCase):
    async def get_application(self):
        app = await super().get_application()

        async def process(self, value):
            return {"result": value + " cow"}

        getMethod = newMethod(
            name="Test Method",
            httpMethod="Get",
            description="This is a test handler",
            process=process,
            urlParameters={
                "value": StringField(),
            },
            responseFields={
                "result": StringField(),
            },
        )

        class end(Endpoint):
            Get = getMethod

        app.addEndpoint(
            "/stuff/{value}/",
            end,
        )
        return app

    @unittest_run_loop
    async def testParameters(self):
        resp = await self.client.request("GET", "/stuff/moo/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()

        self.assertEqual(data["result"], "moo cow")


import inspect
import traceback

from aiohttp import web
from inspect import getdoc

from ..parsing import RequestParser, ResponseSerializer, DictParser
from ..errors import HTTPError, ClientError, ErrorList


class Method():
    name = "Handler"
    description = ""

    def __init__(
        self,
        settings=None,
        statusCode=200,
        headers=None,
        cookies=None,
        deletedCookies=None,
        tests=[],
    ):
        self.statusCode = statusCode
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.deletedCookies = deletedCookies or {}
        self.settings = settings or {}
        self.tests = tests or getattr(self, "tests", [])

    class Parser(RequestParser):
        pass

    class Serializer(ResponseSerializer):
        pass

    async def parse(self, request):
        return await self.Parser(settings=self.settings).parse(request)

    async def process(self, **kwargs):
        return {}

    async def serialize(self, result):
        return await self.Serializer().serialize(
            result,
            statusCode=self.statusCode,
            headers=self.headers,
            cookies=self.cookies,
            deletedCookies=self.deletedCookies,
        )

    def setCookie(
        self,
        name,
        value,
        domain=None,
        secure=False,
        httpOnly=False,
        expires=None,
    ):
        self.cookies[name] = {
            "name": name,
            "value": value,
            "domain": domain,
            "secure": secure,
            "httpOnly": httpOnly,
            "expires": expires,
        }

    def unsetCookie(
        self,
        name,
        domain=None,
    ):
        self.deletedCookies[name] = {
            "name": name,
            "domain": domain,
        }

    def setHeader(self, name, value):
        if name in self.headers:
            self.headers[name].append(value)
        else:
            self.headers[name] = [value]

    async def handle(self, request: web.Request) -> web.Response:
        try:
            parsedData = await self.parse(request)
            parameters = inspect.signature(self.process).parameters
            if "currentUser" in parameters:
                # if the process function wants the current user, fill it
                # in from the authentication middleware
                currentUser = getattr(
                    request,
                    "currentUser",
                    None,
                )
                if all([
                    currentUser == None,
                    parameters["currentUser"].default ==
                    inspect.Parameter.empty,
                ]):
                    raise ClientError(
                        statusCode=401,
                        message="Unauthorized",
                        errorType="AuthenticationError",
                    )

                parsedData["currentUser"] = currentUser
            result = await self.process(**parsedData)
            return await self.serialize(result, )
        except ErrorList as errors:
            return web.json_response(
                {"errors": [e.toJSON() for e in errors.errors]},
                status=errors.errors[0].statusCode,
            )
        except web.HTTPException as err:
            raise err
        except HTTPError as e:
            return web.json_response(
                {"errors": [e.toJSON()]},
                status=e.statusCode,
            )
        except Exception as e:
            print(e)
            print(traceback.format_exc())
            return web.json_response(
                {
                    "errors": [{
                        "message": "Server Error",
                        "errorType": "Unknown",
                    }]
                },
                status=500,
            )

    def describe(self):
        return {
            "name": self.name,
            "description": getattr(self, "description", getdoc(self)),
            "requestParameters": self.Parser().describe(),
            "response": self.Serializer().describe(),
            "examples": [{
                "request": test.request,
                "response": test.response,
            } for test in self.getTests()]
        }

    def getTests(self):
        return self.tests or []


def newMethod(
    name,
    httpMethod,
    description,
    process,
    urlParameters=None,
    queryParameters=None,
    cookieParameters=None,
    headerParameters=None,
    bodyParameters=None,
    responseFields=None,
    tests=None,
):
    """
    Create a new class for the given http method, filling in the
    parsers and the serializer with the given values.
    """
    classMap = {
        "UrlParser": urlParameters,
        "QueryParser": queryParameters,
        "CookieParser": cookieParameters,
        "HeaderParser": headerParameters,
        "BodyParser": bodyParameters,
    }

    parserAttributes = {}
    for key, params in classMap.items():
        if params:
            parserAttributes[key] = type(
                key,
                (DictParser, ),
                params,
            )

    attributes = {
        "name": name,
        "process": process,
        "__doc__": description,
        "description": description,
        "Parser": type(
            "Parser",
            (RequestParser, ),
            parserAttributes,
        ),
        "Serializer": type(
            "Serializer",
            (ResponseSerializer, ),
            responseFields or {},  # Can't be None
        ),
        "tests": tests or [],
    }

    return type(
        httpMethod,
        (Method, ),
        attributes,
    )



from ..parsing import json

from aiohttp.test_utils import AioHTTPTestCase
from dataclasses import dataclass, field

from ..tests.test_case import unittest_run_loop


@dataclass
class TestRequest():
    path: str
    method: str = "GET"
    headers: dict = field(default_factory=dict)
    cookies: dict = field(default_factory=dict)
    body: str = None


@dataclass
class TestResponse():
    statusCode: int
    message: str = ""
    headers: dict = field(default_factory=dict)
    body: str = None


class HttpTest():
    def __init__(
        self,
        request,
        response,
        setUpAsync=None,
        tearDownAsync=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.request = request
        self.response = response
        self.setUpAsync = setUpAsync
        self.tearDownAsync = tearDownAsync

    def generateTestCase(self, appClass, settings):
        expectedRequest = self.request
        expectedResponse = self.response

        theSettings = settings

        class TestCase(AioHTTPTestCase):
            _setUpAsync = self.setUpAsync
            _tearDownAsync = self.tearDownAsync
            settings = theSettings

            async def get_application(self):
                return appClass(settings, logging=False)

            async def compareResponses(self, expected, received):
                if expected.statusCode is not None:
                    self.assertEqual(
                        expectedResponse.statusCode,
                        received.status,
                    )
                if expected.body is not None:
                    receivedBody = await received.text()
                    # check json contents, rather than raw equality
                    if "application/json" in received.headers.get(
                        "Content-Type"
                    ):
                        expectedJson = json.loads(expected.body)
                        receivedJson = json.loads(receivedBody)
                        self.assertDictEqual(expectedJson, receivedJson)
                    else:
                        self.assterEqual(expected.body, receivedBody)

            @unittest_run_loop
            async def testRequest(self):
                if self._setUpAsync is not None:
                    await self._setUpAsync()
                response = await self.client.request(
                    method=expectedRequest.method,
                    path=expectedRequest.path,
                    data=expectedRequest.body,
                    headers=expectedRequest.headers,
                )
                await self.compareResponses(
                    expectedResponse,
                    response,
                )
                return

        return TestCase


from ..handler import Method, Endpoint, newMethod
from ..parsing import (
    RequestParser,
    DictParser,
    ResponseSerializerWithErrors,
)
from ..parsing.fields import (
    NonBlankStringField,
    EmailField,
    StringField,
    UsernameField,
    Field,
    DictField,
    ListField,
)

from ..schemas.user import createUser

from ..schemas.authentication import authenticateUser
from ..middleware import setCurrentUser

from ..errors import ServerError
import jwt
import datetime
import bson
import urllib
from aiohttp import ClientSession, web


class SignUpEndpoint(Endpoint):
    """
        Handle Initial account creation
    """

    name = "Sign Up"

    async def postProcess(
        self,
        email: str,
        username: str,
        password: str,
    ):

        user = await createUser(
            db=self.settings.db,
            username=username,
            email=email,
            password=password,
        )

        token = await setCurrentUser(
            self,
            self.settings.db,
            self.settings.jwtKey,
            user,
        )
        return {"success": "yes"}

    Post = newMethod(
        name="Sign Up",
        httpMethod="POST",
        description=
        "Create a new user with the given username, email, and password.",
        process=postProcess,
        bodyParameters={
            "email": EmailField(
                description="The email of the user being created"
            ),
            "username": UsernameField(
                description="The username of the user being created"
            ),
            "password": NonBlankStringField(
                description="The password to set for the new user"
            ),
        },
        responseFields={
            "success": Field(
                description="A boolean of whether the user was created"
            ),
            "errors": ListField(DictField({})),
        }
    )


class LogInEndpoint(Endpoint):
    """
    Log in a user
    """
    name = "Log In"

    async def postProcess(
        self,
        email: str,
        password: str,
    ):
        user = await authenticateUser(
            db=self.settings.db,
            email=email,
            password=password,
        )
        token = await setCurrentUser(
            self,
            self.settings.db,
            self.settings.jwtKey,
            user,
        )

        return {
            "token": token,
            "currentUser": {
                "id": str(user["_id"]),
                "username": user["username"],
                "email": user["email"],
            },
        }

    Post = newMethod(
        name="Log In",
        httpMethod="POST",
        description="Log in the use with the given email and password",
        process=postProcess,
        bodyParameters={
            "email": EmailField(),
            "password": StringField(),
        },
        responseFields={
            "token": StringField(),
            "currentUser": DictField({
                "id": StringField(),
                "username": StringField(),
                "email": EmailField(),
            }),
            "errors": ListField(DictField({})),
        }
    )


class LogOutEndpoint(Endpoint):
    """
    Log out the Current User
    """

    name = "Log Out"

    async def process(self, ):
        self.unsetCookie("authentication", )
        self.unsetCookie("currentUser", )
        return {}

    Get = newMethod(
        name="Log Out",
        httpMethod="GET",
        description=
        "Log out the current user by unsetting the relevant cookies",
        process=process,
        responseFields={
            "errors": ListField(DictField({})),
        }
    )


class MeEndpoint(Endpoint):
    """
    Return information about the currently logged in user
    """

    name = "Me"

    async def processGet(self, currentUser):
        await setCurrentUser(
            self,
            self.settings.db,
            self.settings.jwtKey,
            currentUser,
        )
        return {
            "currentUser": {
                "id": str(currentUser["_id"]),
                "username": currentUser["username"],
                "email": currentUser["email"],
            },
        }

    Get = newMethod(
        name="Get Current User",
        httpMethod="GET",
        description="Return information about the currently logged in user",
        process=processGet,
        responseFields={
            "currentUser": DictField({
                "id": StringField(),
                "username": StringField(),
                "email": EmailField(),
            }),
            "errors": ListField(DictField({})),
        }
    )


def generateOAuthEndpoint(
    clientID: str,
    clientSecret: str,
    redirectUri: str,
    responseType: str,
    grantType: str,
    scope: str,
    authorizationUri: str,
    tokenUri: str,
    handleTokenResponse=None,
):
    async def process(
        self,
        code=None,
        state=None,
    ):
        if code == None:
            raise web.HTTPFound(
                location=authorizationUri + "?" + urllib.parse.urlencode({
                    "response_type": responseType,
                    "client_id": clientID,
                    "redirect_uri": redirectUri,
                    "scope": scope,
                })
            )

        async with ClientSession() as session:
            async with session.get(
                tokenUri,
                params={
                    "code": code,
                    "grant_type": grantType,
                    "redirect_uri": redirectUri,
                    "client_id": clientID,
                    "client_secret": clientSecret,
                }
            ) as response:
                token = await response.json()
                if handleTokenResponse:
                    await handleTokenResponse(
                        handler=self,
                        token=token,
                    )

    class OAuthEndpoint(Endpoint):
        Get = newMethod(
            name="Oauth Login",
            httpMethod="GET",
            description="OAuth 2.0 login endpoint",
            process=process,
            queryParameters={
                "code": StringField(optional=True),
                "state": StringField(optional=True),
            }
        )

    return OAuthEndpoint



from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict
import inspect


@dataclass
class DictField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.OBJECT])

    def __init__(self, fields: dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fields = fields or {}
        for k, v in fields.items():
            setattr(self, k, v)

    def _getFields(self):
        return {
            f: theField
            for (f, theField) in filter(
                lambda k: isinstance(k[1], Field),
                inspect.getmembers(self),
            )
        }

    def parser(self, source: Dict):
        source = source or {}
        fields = self._getFields()
        if not fields:
            return source
        parameters = {}
        for name, parameter in fields.items():
            if (not parameter.optional) and (name not in source):
                raise ParseError(
                    message="Missing required field: {}".format(name),
                    statusCode=parameter.missingStatusCode,
                )
            parameters[name] = parameter.parse(
                source.get(name, parameter.default)
            )
        return parameters

    def serialize(self, dictValue):
        fields = self._getFields()
        if not fields:
            return dictValue
        result = {}

        for name, outField in fields.items():
            try:
                value = dictValue[name]
                name = getattr(outField, "serializeTo", name) or name

                try:
                    result[name] = outField.serialize(value)
                except:
                    result[name] = value
            except KeyError as k:
                if not outField.optional:
                    result[name] = None

        return result

    def describe(self):
        return {
            key: field.describe()
            for (key, field) in self._getFields().items()
        }



from .parsing import json
import markdown

from aiohttp import web
from aiohttp.web_middlewares import normalize_path_middleware
from types import SimpleNamespace as SN
from typing import Sequence

from .routes import (
    authentication as authenticationRoutes,
    permissions as permissionRoutes,
    health as healthRoutes,
)
from .config import defaultSettings
from .middleware import (
    logger,
    allowCors,
    removeServerHeader,
    authentication as authenticationMiddleware,
)
from .handler import (
    newMethod,
    Endpoint,
    Method,
)

from .parsing import handleAcceptHeader
from .documentation import requestDocumentationToApiBlueprint


class Clericus(web.Application):
    def __init__(self, settings=None, logging=True, usernameField="username"):
        baseSettings = defaultSettings()
        baseSettings.update(settings or {})
        middlewares = [
            normalize_path_middleware(append_slash=True),
            allowCors(origins=baseSettings["corsOrigins"]),
            removeServerHeader,
        ]

        if logging:
            middlewares.append(logger(usernameField))

        middlewares.append(
            authenticationMiddleware(
                db=baseSettings["db"],
                secretKey=baseSettings["jwtKey"],
            ),
        )
        super().__init__(middlewares=middlewares, )

        self.documentation = []
        self.tests = []

        self["settings"] = baseSettings

        self.router.add_route(
            "GET",
            "/",
            self.documentationHandler,
        )

        self.addEndpoint(
            "/healthy/",
            healthRoutes.HealthCheckEndpoint,
            name="Health Check",
        )

    def addEndpoint(self, path, handlerClass, name=None):
        cls = handlerClass(
            settings=SN(**self["settings"]),
            name=name,
            path=path,
        )
        self.router.add_route("*", path, cls.handle)
        self.router.add_route(
            "get",
            f"/documentation{path}",
            cls.handleDocumentation,
        )

        self.documentation.append(cls.describe())
        self.tests += cls.getTests()

    async def documentationHandler(self, request: web.Request) -> web.Response:
        docs = self.describe()

        acceptHeader = request.headers.get("Accept")

        if acceptHeader:
            contentType, _ = handleAcceptHeader(
                acceptHeader,
                [
                    "application/json",
                    "text/html",
                ],
            )
        else:
            contentType = "application/json"

        if contentType == "text/html":
            md = "# Documentation\n"
            for endpoint in docs["endpoints"]:
                md += requestDocumentationToApiBlueprint(endpoint)
            htmlContent = markdown.markdown(md)
            html = f"<html><body>{htmlContent}</body></html>"

            return web.Response(
                text=html,
                headers={"Content-Type": "text/html"},
            )

        else:
            return web.Response(
                text=json.dumps(docs),
                headers={"Content-Type": "application/json"},
            )

    def describe(self):
        return {
            "endpoints": sorted(
                self.documentation,
                key=lambda k: k["path"],
            )
        }

    def getTests(self):
        return {
            f"test{n+1}": t.generateTestCase(
                settings=self["settings"],
                appClass=self.__class__,
            )
            for n, t in enumerate(self.tests)
        }

    def runApp(self):
        web.run_app(self)



import unittest
import asyncio

from ...parsing import DictParser, handleAcceptHeader
from ...parsing.fields import StringField, DictField
from ..test_case import async_test


class TestParsing(unittest.TestCase):
    @async_test
    async def test_moo(self):
        # self.assertEqual(1, 1)
        class dp(DictParser):
            stringOne = StringField()
            stringTwo = StringField()
            dictOne = DictField(
                fields={
                    "stringThree": StringField(),
                    "stringFour": StringField(),
                    "dictTwo": DictField({}),
                }
            )

        parsed = await dp().parse({
            "stringOne": "moo",
            "stringTwo": "cow",
            "dictOne": {
                "stringThree": "moo",
                "stringFour": "cow",
                "dictTwo": {
                    "x": 1
                }
            }
        })

        self.assertEqual(parsed["stringOne"], "moo")
        self.assertEqual(parsed["stringTwo"], "cow")
        self.assertEqual(parsed["dictOne"]["stringThree"], "moo")
        self.assertEqual(parsed["dictOne"]["stringFour"], "cow")
        self.assertEqual(parsed["dictOne"]["dictTwo"]["x"], 1)


class TestAcceptHeaderParsing(unittest.TestCase):
    def test_simple_parsing(self):
        contentType, parameters = handleAcceptHeader(
            "text/*;q=0.3, text/html;q=0.7, text/html;level=1, text/html;level=2;q=0.4, */*;q=0.5",
            ["text/html", "text/plain"],
        )
        self.assertEqual(contentType, "text/html")
        self.assertEqual(parameters["level"], "1")


if __name__ == '__main__':
    unittest.main()


import math

from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict


@dataclass
class IntegerField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.INTEGER])
    minimum: int = None
    maximum: int = None

    def parser(self, value):
        value = super().parser(value)
        if isinstance(value, str):
            try:
                value = int(value)
            except:
                raise ParseError(message=f"\"{value}\" is not an integer")
        if not isinstance(value, int):
            raise ParseError(message=f"\"{value}\" is not an integer")
        if self.minimum is not None and value < self.minimum:
            raise ParseError(message=f"{value} is less than {self.minimum}")
        if self.maximum is not None and value > self.maximum:
            raise ParseError(message=f"{value} is greater than {self.maximum}")
        return value


@dataclass
class FloatField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.FLOAT])
    allowNAN: bool = False
    allowInf: bool = False
    minimum: float = None
    maximum: float = None

    def parser(self, value):
        value = super().parser(value)
        if isinstance(value, str):
            try:
                value = float(value)
            except:
                raise ParseError(message=f"\"{value}\" is not an float")
        if not isinstance(value, float):
            raise ParseError(message=f"\"{value}\" is not an float")
        if not self.allowNAN and math.isnan(value):
            raise ParseError(message="NAN is not allowed")
        if not self.allowNAN and math.isinf(value):
            raise ParseError(message="Infinity is not allowed")
        if self.minimum is not None and value < self.minimum:
            raise ParseError(message=f"{value} is less than {self.minimum}")
        if self.maximum is not None and value > self.maximum:
            raise ParseError(message=f"{value} is greater than {self.maximum}")

        return value


import unittest
import faker

from aiohttp.test_utils import unittest_run_loop

from ..test_case import ClericusTestCase
from ...routes.authentication import generateOAuthEndpoint

fake = faker.Faker()


class OAuthTestCase(ClericusTestCase):
    async def get_application(self):
        app = await super().get_application()

        return app

    @unittest_run_loop
    async def testOauth(self):

        # print(self.app)

        resp = await self.client.request("GET", "/oauth/test/")
        # self.assertEqual(resp.status, 401)
        # data = await resp.json()
        # print(resp)
        # print(data)


from aiohttp.web import middleware, HTTPError
import datetime
import bson
import jwt
from colors import green, red


def logger(usernameField="username"):
    @middleware
    async def logRequest(request, handler):
        start = datetime.datetime.utcnow()
        try:
            response = await handler(request)

            status = (
                red(response.status)
                if response.status >= 400 else green(response.status)
            )
            print(
                status,
                request.method,
                request.path,
                getattr(
                    request,
                    "currentUser",
                    {usernameField: "Unauthenticated"},
                )[usernameField],
                f"{int((datetime.datetime.utcnow() - start).total_seconds() * 1000)}ms",
            )

            return response
        except HTTPError as err:
            print(
                red(err.status),
                request.method,
                request.path,
                getattr(
                    request,
                    "currentUser",
                    {usernameField: "Unauthenticated"},
                )[usernameField],
                f"{int((datetime.datetime.utcnow() - start).total_seconds() * 1000)}ms",
            )

            raise err
        except Exception as err:
            print(type(err), err)
            raise err

    return logRequest


def allowCors(origins):
    @middleware
    async def middle(request, handler):
        resp = await handler(request)
        try:
            origin = request.headers.get('Origin')[0]
        except:
            origin = None
        resp.headers["Access-Control-Allow-Credentials"] = "true"
        if origin in origins:
            resp.headers.add("Access-Control-Allow-Origin", origin)
        else:
            defaultOrigin = origins[0] if origins else "*"
            resp.headers.add("Access-Control-Allow-Origin", defaultOrigin)

        try:
            requestHeaders = request.headers.get(
                'Access-Control-Request-Headers'
            )
            if requestHeaders:
                resp.headers.add(
                    "Access-Control-Allow-Headers", requestHeaders
                )
        except:
            pass

        return resp

    return middle


@middleware
async def removeServerHeader(request, handler):
    response = await handler(request)
    response.headers["Server"] = ""
    return response


def authentication(
    db, secretKey, cookieName="authentication", headerKey="X-Auth-Token"
):
    @middleware
    async def middle(request, handler):
        try:
            value = request.cookies[cookieName]
        except:
            value = None

        if value is None and headerKey:
            value = request.headers.get(headerKey, None)

        if value:
            try:
                parsed = jwt.decode(
                    value,
                    secretKey,
                    algorithms=['HS256'],
                )
                request.currentUser = await db.user.find_one({
                    "_id": bson.ObjectId(parsed["id"]),
                })
            except:
                # TODO: Log this
                pass
        return await handler(request)

    return middle


async def setCurrentUser(
    method,
    db,
    secretKey,
    user,
    cookiename="authentication",
    expiration=datetime.timedelta(days=1)
):
    now = datetime.datetime.utcnow()
    token = jwt.encode(
        payload={
            "id": str(user["_id"]),
            "username": user["username"],
            "email": user["email"],
            "exp": now + expiration,
            "iat": now,
        },
        key=secretKey,
        algorithm='HS256',
    ).decode("utf-8")
    method.setCookie("authentication", token)
    return token


import unittest
import faker

from aiohttp.test_utils import unittest_run_loop

from ..test_case import ClericusTestCase

fake = faker.Faker()


class HealthCheckTestCase(ClericusTestCase):
    @unittest_run_loop
    async def testHealthCheck(self):
        resp = await self.client.request("GET", "/healthy/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()

        self.assertEqual(data["healthy"], True)


from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict

from dateutil.parser import parse as parseDate


@dataclass
class DatetimeField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.STRING])

    def parser(self, value):
        try:
            return parseDate(value)
        except:
            raise ParseError(
                message="\"{}\" is not a valid datetime".format(value)
            )
        return value

    def serialize(self, value):
        if value:
            return value.isoformat()
        return None


import dataclasses
from json import *
from json import dumps as _dumps


class EnhancedJSONEncoder(JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


def dumps(o, *args, **kwargs):
    return _dumps(o, cls=EnhancedJSONEncoder, *args, **kwargs)


class HTTPError(Exception):
    def __init__(
        self,
        message="Server Error",
        statusCode=500,
        errorType=None,
        errorCode=None,
        key=None,
    ):
        self.statusCode = statusCode
        self.message = message
        self.errorType = errorType
        self.errorCode = errorCode
        self.key = key

    def __repr__(self):
        return self.message

    def toJSON(self):
        d = {
            "message": self.message,
        }

        for k in ["errorType", "errorCode", "key"]:
            if getattr(self, k, None) != None:
                d[k] = getattr(self, k)

        return d


class ClientError(HTTPError):
    def __init__(
        self,
        message="Invalid Request",
        statusCode=400,
        errorType=None,
        errorCode=None,
        key=None,
    ):
        super().__init__(
            message,
            statusCode,
            errorType,
            errorCode,
            key,
        )


class ServerError(HTTPError):
    def __init__(
        self,
        message="Server Error",
        statusCode=500,
        errorType=None,
        errorCode=None,
        key=None,
    ):
        super().__init__(
            message,
            statusCode,
            errorType,
            errorCode,
            key,
        )


class ErrorList(Exception):
    def __init__(self, errors):
        self.errors = errors


from ..parsing import json
import copy
from types import SimpleNamespace
import motor.motor_asyncio


def defaultSettings():
    return {
        "db": {
            "url": "localhost:27017",
            "name": "test",
        },
        "corsOrigins": [
            "http://localhost:3000",
            # "http://localhost:8080",
        ],
        "jwtKey": "onetwothree",
    }


def parseConfig(filename="config.json", default=defaultSettings):
    settings = copy.deepcopy(default())
    if filename:
        settings.update(json.loads(open(filename).read()))

    return settings


def connectToDB(settings):
    client = motor.motor_asyncio.AsyncIOMotorClient(
        settings["db"]["url"],
        tz_aware=True,
    )
    settings["db"] = client[settings["db"]["name"]]
    return settings



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="clericus",
    version="0.0.3a24",
    author="Joseph L Buell V",
    author_email="jlrbuellv@gmail.com",
    description=
    "An async webserver focused on being predictable and self documenting.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mrincredibuell/clericus",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=[
        "aiohttp>=3.5.4",
        "pyjwt>=1.7.1",
        "motor>=2.0.0",
        "python-dateutil>=2.8.0",
        "bcrypt>=3.1.6",
        "dnspython>=1.16.0",
        "faker>=1.0.7",
        "markdown>=3.1.1",
        "ansicolors>=1.1.8",
    ],
)


import bson

from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict


@dataclass
class ObjectIdField(Field):
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.STRING])

    def parser(self, value):
        try:
            return bson.ObjectId(value)
        except:
            raise ParseError(message="\"{}\" is not a valid id".format(value))
        return value

    def serialize(self, value):
        return str(value)



from ..handler import Method, Endpoint, newMethod
from ..parsing import (
    RequestParser,
    DictParser,
    ResponseSerializer,
)
from ..parsing.fields import (
    NonBlankStringField,
    EmailField,
    StringField,
    JwtField,
    BoolField,
    Field,
)

from ..errors import ServerError
import jwt
import datetime


class HealthCheckEndpoint(Endpoint):
    """
        Return the status of the server
    """
    name = "Health Check"

    async def getProcess(self, ):
        try:
            resp = await self.settings.db.command("ping")

            return {"healthy": True}
        except Exception as e:
            print(e)
            raise ServerError(message="Server is unhealthy")

    Get = newMethod(
        name="Health Check",
        httpMethod="Get",
        description="""Return the status of the server""",
        process=getProcess,
        responseFields={
            "healthy": Field(
                description="A boolean of whether the server is healthy",
                default=True,
            )
        }
    )



import string

from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict


@dataclass
class StringField(Field):
    """
    A generic string parameter
    """
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.STRING])

    def parser(self, value):
        value = super().parser(value)
        if not isinstance(value, str):
            raise ParseError(message="\"{}\" is not a string".format(value))
        return value


@dataclass
class EnumeratedStringField(Field):
    """
    A string parameter which must be one of a particular set of values
    """
    allowedTypes: List = field(default_factory=lambda: [FieldTypes.STRING])
    allowedValues: List = field(default_factory=lambda: [])

    def parser(self, value):
        originalValue = value
        value = super().parser(value)
        if value not in self.allowedValues:
            raise ParseError(
                message=
                f"\"{originalValue}\" is not in {str(self.allowedValues)}"
            )
        return value

    def describe(self):
        doc = super().describe()
        doc["allowedValues"] = self.allowedValues
        return doc


@dataclass
class NoWhitespaceStringField(StringField):
    """
    A string field where leading/trailing whitespace
    should be trimmed.
    """

    def normalizer(self, value):
        value = super().normalizer(value)
        return value.strip()


@dataclass
class NonBlankStringField(NoWhitespaceStringField):
    def parser(self, value):
        value = super().parser(value)
        if not value:
            raise ParseError(message="Cannot be blank")
        return value


@dataclass
class EmailField(NonBlankStringField):
    def parser(self, value):
        value = super().parser(value)
        if "@" not in value:
            raise ParseError(message=f"{value} is not a valid email address", )
        return value


@dataclass
class UsernameField(NonBlankStringField):
    def parser(self, value):
        value = super().parser(value)

        if any([c in value for c in string.whitespace]):
            raise ParseError("Usernames must not contain whitespace")

        return value



import unittest
import asyncio
from aiohttp.test_utils import make_mocked_request

from ...handler import newMethod, Endpoint
from ...handler.testing import HttpTest, TestRequest, TestResponse
from ...parsing import json
from ...parsing.fields import BoolField, StringField, EnumeratedStringField

from ..test_case import (
    ClericusTestCase,
    unittest_run_loop,
    async_test,
)

from ... import Clericus
from ...config import defaultSettings, connectToDB

settings = defaultSettings()
settings["db"]["name"] = f"test{'moo'}"
settings = connectToDB(settings)


# await settings["db"].client.drop_database(settings["db"].name)
# await createCollections(self.db)
async def process(self, exampleValue):
    return {"result": exampleValue + " cow"}


getMethod = newMethod(
    name="Test GET Method",
    httpMethod="Get",
    description="This is a test handler",
    process=process,
    urlParameters={
        "exampleValue": StringField(description="A string to modify", ),
    },
    responseFields={
        "result": StringField(description="The value with \"cow\" appended", ),
    },
    tests=[
        HttpTest(
            request=TestRequest(path="/stuff/moo", ),
            response=TestResponse(
                statusCode=200,
                body=json.dumps({"result": "moo cow"}, indent=1),
            ),
        )
    ]
)


class end(Endpoint):
    """
    An example endpoint.
    """
    Get = getMethod


class myServer(Clericus):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.addEndpoint(
            "/stuff/{exampleValue}/",
            end,
            name="Example Endpoint",
        )


app = myServer(settings, logging=False)

globals().update(app.getTests())



from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict


@dataclass
class ListField(Field):
    def __init__(self, elementField):
        super().__init__()
        self.elementField = elementField
        self.default = []

    def parser(self, source):
        if not source:
            return []
        parsed = []
        for value in source:
            parsed.append(elementField.parser(value))
        return parsed

    def serialize(self, values):
        result = []
        if not values:
            return result
        for value in values:
            try:
                result.append(self.elementField.serialize(value))
            except Exception as e:
                print(e)
                result.append(value)

        return result


from .fields import Field, FieldTypes
from ..errors import ParseError

from dataclasses import dataclass, field
from typing import Any, List, Dict

from .strings import StringField
from .dicts import DictField


@dataclass
class ErrorField(DictField):
    message = StringField()



import unittest
import asyncio

from aiohttp.test_utils import make_mocked_request

from ...parsing import json

from ...handler import newMethod, Endpoint
from ...parsing.fields import BoolField, StringField, EnumeratedStringField
from ...handler.testing import HttpTest, TestRequest, TestResponse

from ..test_case import (
    ClericusTestCase,
    unittest_run_loop,
    async_test,
)


class TestDocumentation(ClericusTestCase):
    async def get_application(self):
        app = await super().get_application()

        async def process(self, exampleValue, q=""):
            return {"result": exampleValue + " cow" + q}

        getMethod = newMethod(
            name="Test GET Method",
            httpMethod="Get",
            description="This is a test handler",
            process=process,
            urlParameters={
                "exampleValue": StringField(
                    description="A string to modify",
                ),
            },
            queryParameters={
                "q": StringField(
                    description="A query parameter",
                    optional=True,
                ),
                "moo": StringField(
                    description="Another query parameter",
                    optional=False,
                    default="moo",
                ),
            },
            responseFields={
                "result": StringField(
                    description="The value with \"cow\" appended",
                ),
            },
            tests=[
                HttpTest(
                    request=TestRequest(path="/stuff/moo?q=!", ),
                    response=TestResponse(
                        statusCode=200,
                        body=json.dumps({"result": "moo cow!"}, indent=2),
                    ),
                )
            ]
        )

        postMethod = newMethod(
            name="Test POST Method",
            httpMethod="Post",
            description="This is a test post handler",
            process=process,
            urlParameters={
                "exampleValue": StringField(
                    description="A string to modify",
                ),
            },
            bodyParameters={
                "q": StringField(
                    description="A body parameter",
                    optional=True,
                ),
                "animal": EnumeratedStringField(
                    description="An enumerated parameter",
                    optional=False,
                    default="cow",
                    allowedValues=[
                        "cow",
                        "chicken",
                        "pig",
                    ]
                ),
            },
            responseFields={
                "result": StringField(
                    description="The value with \"cow\" appended",
                ),
            },
        )

        class end(Endpoint):
            """
            An example endpoint.
            """
            Get = getMethod
            Post = postMethod

        app.addEndpoint(
            "/stuff/{exampleValue}/",
            end,
            name="Example Endpoint",
        )
        return app

    @unittest_run_loop
    async def testBaseDocumentationJson(self):
        resp = await self.client.request("GET", "/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        endpoints = data["endpoints"]
        self.assertGreater(len(endpoints), 1)

        data = next(
            filter(lambda k: k["path"] == "/stuff/{exampleValue}/", endpoints)
        )
        self.assertEqual(data["description"], "An example endpoint.")
        self.assertEqual(data["name"], "Example Endpoint")
        self.assertEqual(data["path"], "/stuff/{exampleValue}/")
        self.assertEqual(
            data["methods"]["get"]["description"],
            "This is a test handler",
        )

        self.assertEqual(
            data["methods"]["get"]["requestParameters"]["url"]["exampleValue"]
            ["description"],
            "A string to modify",
        )

    @unittest_run_loop
    async def testBaseDocumentationHtml(self):
        resp = await self.client.request(
            "GET",
            "/",
            headers={
                "Accept": "text/html",
            },
        )
        self.assertEqual(resp.status, 200)
        data = await resp.text()

    @unittest_run_loop
    async def testEndpointDocumentationJson(self):
        resp = await self.client.request("GET", "/documentation/stuff/moo/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()

        self.assertEqual(data["description"], "An example endpoint.")
        self.assertEqual(data["name"], "Example Endpoint")
        self.assertEqual(data["path"], "/stuff/{exampleValue}/")
        self.assertEqual(
            data["methods"]["get"]["description"],
            "This is a test handler",
        )

        self.assertEqual(
            data["methods"]["get"]["requestParameters"]["url"]["exampleValue"]
            ["description"],
            "A string to modify",
        )

    @unittest_run_loop
    async def testEndpointDocumentationHtml(self):
        resp = await self.client.request(
            "GET",
            "/documentation/stuff/moo/",
            headers={
                "Accept": "text/html",
            },
        )
        self.assertEqual(resp.status, 200)
        data = await resp.text()



from .fields import Field, JwtField

from .strings import (
    StringField,
    EnumeratedStringField,
    NoWhitespaceStringField,
    NonBlankStringField,
    EmailField,
    UsernameField,
)

from .numbers import IntegerField, FloatField
from .booleans import BoolField
from .dates import DatetimeField
from .lists import ListField
from .dicts import DictField
from .mongo import ObjectIdField
from .errors import ErrorField
